/*
 * Copyright (c) 2011, Red Hat, Inc.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RED HAT, INC. DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RED HAT, INC. BE LIABLE
 * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Author: Jan Friesse <jfriesse@redhat.com>
 */

#ifndef _GCRA_H_
#define _GCRA_H_

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Structures definition
 */
struct gcra_item {
	uint64_t tat;
	unsigned int interval;
	unsigned int tau;
};

/*
 * Prototypes
 */
extern void		gcra_init(struct gcra_item *item, unsigned int interval,
    unsigned int burst);

extern int		gcra_rl(struct gcra_item *item, struct timeval tv);

#ifdef __cplusplus
}
#endif

#endif /* _GCRA_H_ */
